@@ -75,7 +75,7 @@ class AStarResults:
 
 
 class Edge(object):
-    def __init__(self, nodeA: Node, nodeB: Node, edge_direction=None):
+    def __init__(self, nodeA: Node, nodeB: Node, edge_direction=None, naming_provider: Callable[[], str] = None):
         self.start = nodeA
         self.end = nodeB
         if edge_direction is None:
@@ -83,7 +83,7 @@ class Edge(object):
         self.direction = edge_direction
         self._disablers = set()
         self.length = nodeA.pos.distance_from(nodeB.pos)
-        self.id = str(uuid.uuid4())
+        self.id = naming_provider() if naming_provider else str(uuid.uuid4())
 
     def __str__(self):
         char_dict = {
@@ -136,7 +136,52 @@ class Edge(object):
 
 class Graph(object):
 
-    def __init__(self, graph_dict: Dict[Node, List[Node]]=None):
+    #TODO: Add a "from file" constructor cls method
+
+    # def create_points_from_file(filePath: str, scale: (int, int), hasHeaders: bool = True):
+    #     import csv
+    #
+    #     points = []
+    #     with open(filePath) as csv_file:
+    #         csv_reader = csv.reader(csv_file, delimiter=',')
+    #         line_count = 0
+    #
+    #         for row in csv_reader:
+    #             if line_count == 0 and hasHeaders:
+    #                 print(f'Column names are {", ".join(row)}')
+    #                 line_count += 1
+    #             else:
+    #                 # end_node= Node(str(row[0]) + "_start", int(float(row[1])), int(float(row[3]))
+    #
+    #                 points.append((int(float(row[1]) * scale[0]), int(float(row[3]) * scale[1])))
+    #                 line_count += 1
+    #
+    #     return points
+    #
+    # def create_edges_from_file(filePath: str, scale: (int, int), hasHeaders: bool = True):
+    #     import csv
+    #
+    #     edges = []
+    #     with open(filePath) as csv_file:
+    #         csv_reader = csv.reader(csv_file, delimiter=',')
+    #         line_count = 0
+    #
+    #         for row in csv_reader:
+    #             if line_count == 0 and hasHeaders:
+    #                 print(f'Column names are {", ".join(row)}')
+    #                 line_count += 1
+    #             else:
+    #                 start_node = Node(str(row[0]) + "_start", int(float(row[1]) * scale[0]),
+    #                                   int(float(row[2]) * scale[1]))
+    #                 end_node = Node(str(row[0]) + "_end", int(float(row[3]) * scale[0]), int(float(row[4]) * scale[1]))
+    #                 edges.append(Edge(row[0], start_node, end_node, EdgeConnection.TwoWay))
+    #                 line_count += 1
+    #
+    #     return edges
+
+
+
+    def __init__(self, graph_dict: Dict[Node, List[Node]]=None, naming_provider: Callable[[], str] = None):
         """ initializes a graph object
             If no dictionary or None is given, an empty dictionary will be used
         """
@@ -144,10 +189,10 @@ class Graph(object):
             graph_dict = {}
         self._graph_dict = graph_dict
 
+        self.naming_provider = naming_provider
 
-
-        self._nodes = {}
-        self._edges = {}
+        self._nodes_dict = {}
+        self._edges_dict = {}
 
         ''' self._edges[start: Vector2][end: Vector2] = edge: Edge '''
         self._pos_edge_map = None
@@ -160,23 +205,23 @@ class Graph(object):
 
 
         for node in graph_dict.keys():
-            self._nodes[node.name] = node
+            self._nodes_dict[node.name] = node
             for toNode in graph_dict[node]:
-                edge = Edge(node, toNode, EdgeDirection.ONEWAY)
-                self._edges[edge.id] = edge
+                edge = Edge(node, toNode, EdgeDirection.ONEWAY, naming_provider=self.naming_provider)
+                self._edges_dict[edge.id] = edge
 
         self._build_maps()
 
 
     def _build_maps(self):
-        # self.__graph_dict is a first class citizen. Therefore update all the others off of an updated graph_dict
-        self._graph_dict = self.__generate_graph_dict(self._edges)
+        # self._graph_dict is a first class citizen. Therefore update all the others off of an updated graph_dict
+        self._graph_dict = self.__generate_graph_dict(self._edges_dict)
 
         # update remaining maps assuming graph_dict is accurate
-        self._pos_edge_map = self.__generate_pos_edge_map(self._edges)
-        self._node_edge_map = self.__generate_node_edge_map(self._edges)
-        self._pos_node_map = self.__generate_position_node_map(self._nodes)
-        self._node_by_name_map = self.__generate_node_by_name_map(self._nodes)
+        self._pos_edge_map = self.__generate_pos_edge_map(self._edges_dict)
+        self._node_edge_map = self.__generate_node_edge_map(self._edges_dict)
+        self._pos_node_map = self.__generate_position_node_map(self._nodes_dict)
+        self._node_by_name_map = self.__generate_node_by_name_map(self._nodes_dict)
 
     def __generate_node_edge_map(self, edges: Dict[str, Edge]):
         node_edge_map = {}
@@ -216,7 +261,7 @@ class Graph(object):
         for id, edge in edges.items():
             graph_dict.setdefault(edge.start, []).append(edge.end)
 
-        for id, node in self._nodes.items():
+        for id, node in self._nodes_dict.items():
             if node not in graph_dict.keys():
                 graph_dict[node] = []
         return graph_dict
@@ -227,15 +272,27 @@ class Graph(object):
             ret[node.name] = node_id
         return ret
 
-    def nodes(self):
+    def _nodes(self) -> List[Node]:
         """ returns the vertices of a graph """
         # return copy.deepcopy([node for id, node in self._nodes.items()])
-        return ([node for id, node in self._nodes.items()])
+        return ([node for id, node in self._nodes_dict.items()])
 
-    def edges(self):
+    @property
+    def nodes(self) -> List[Node]:
+        return self._nodes()
+
+    def _edges(self, ids: List[str] = None) -> List[Edge]:
         """ returns the edges of a graph """
         # return copy.deepcopy([edge for id, edge in self._edges.items()])
-        return ([edge for id, edge in self._edges.items()])
+        return ([edge for id, edge in self._edges_dict.items() if ids is None or id in ids])
+
+    @property
+    def edges(self) -> List[Edge]:
+        return self._edges()
+
+    def edges_by_id(self, edge_ids: List[str]):
+        return self._edges(edge_ids)
+
 
     def add_node_with_connnections(self, node: Node, connections: Dict[Node, EdgeDirection]):
         self.add_node(node)
@@ -249,12 +306,12 @@ class Graph(object):
 
         for connection, direction in connections.items():
             if direction == EdgeDirection.TWOWAY:
-                edges.append(Edge(node, connection))
-                edges.append(Edge(connection, node))
+                edges.append(Edge(node, connection, naming_provider=self.naming_provider))
+                edges.append(Edge(connection, node, naming_provider=self.naming_provider))
             elif direction == EdgeDirection.TO:
-                edges.append(Edge(node, connection))
+                edges.append(Edge(node, connection, naming_provider=self.naming_provider))
             elif direction == EdgeDirection.FROM:
-                edges.append(Edge(connection, node))
+                edges.append(Edge(connection, node, naming_provider=self.naming_provider))
 
         self.add_edges(edges)
 
@@ -264,8 +321,8 @@ class Graph(object):
             list as a value is added to the dictionary.
             Otherwise nothing has to be done.
         """
-        if node.name not in self._nodes.keys():
-            self._nodes[node.name] = node
+        if node.name not in self._nodes_dict.keys():
+            self._nodes_dict[node.name] = node
         self._build_maps()
 
     def add_edges(self, edges):
@@ -278,10 +335,10 @@ class Graph(object):
         elif isinstance(edges, dict) and isinstance(list(edges.keys())[0], IVector):
             for start in edges.keys():
                 for end in edges[start]:
-                    start_node = self.nodes_at_point(start)
-                    end_node = self.nodes_at_point(end)
+                    start_node = self.nodes_at_point(start)[0]
+                    end_node = self.nodes_at_point(end)[0]
                     if start_node and end_node:
-                        edge = Edge(start_node, end_node)
+                        edge = Edge(start_node, end_node, naming_provider=self.naming_provider)
                         self._add_edge(edge)
         elif isinstance(edges, Edge):
             self._add_edge(edges)
@@ -297,8 +354,8 @@ class Graph(object):
         elif isinstance(edges, dict) and isinstance(list(edges.keys())[0], IVector):
             for start in edges.keys():
                 for end in edges[start]:
-                    start_node = self.nodes_at_point(start)
-                    end_node = self.nodes_at_point(end)
+                    start_node = self.nodes_at_point(start)[0]
+                    end_node = self.nodes_at_point(end)[0]
                     if start_node and end_node:
                         edge = self._edge_at(start_node.pos, end_node.pos)
                         self._remove_edge(edge)
@@ -343,7 +400,7 @@ class Graph(object):
         if not isinstance(node, Node):
             raise Exception(f"input must be of type {Node}, but {type(node)} was provided")
         node_edges = self._node_edge_map[node]
-        return [self._edges[edge_id] for edge_id in node_edges]
+        return [self._edges_dict[edge_id] for edge_id in node_edges]
 
 
     def disable_edges_to_node(self, node: Node, disabler):
@@ -352,7 +409,7 @@ class Graph(object):
         if isinstance(node, Node):
             node_edges = self._node_edge_map[node]
             for edge_id in node_edges:
-                edge = self._edges[edge_id]
+                edge = self._edges_dict[edge_id]
                 edge.add_disabler(disabler)
 
     def enable_edges_to_node(self, node: Node, disabler):
@@ -361,7 +418,7 @@ class Graph(object):
         if isinstance(node, Node):
             node_edges = self._node_edge_map[node]
             for edge_id in node_edges:
-                edge = self._edges[edge_id]
+                edge = self._edges_dict[edge_id]
                 edge.remove_disabler(disabler)
 
     def adjacent_nodes(self, node: Node, only_enabled: bool = False, ignored_disablers: set = None) -> List[Node]:
@@ -374,19 +431,19 @@ class Graph(object):
     def _remove_edge(self, edge: Edge):
         edge = self._edge_at(edge.start.pos, edge.end.pos)
         if edge.id in self._pos_edge_map.keys():
-            del self._edges[edge.id]
+            del self._edges_dict[edge.id]
         self._build_maps()
 
     def _add_edge(self, edge: Edge):
         existing_edge = self._edge_at(edge.start.pos, edge.end.pos)
         if existing_edge is None:
-            self._edges[edge.id] = edge
+            self._edges_dict[edge.id] = edge
         self._build_maps()
 
     def nodes_at_point(self, pos: IVector) -> List[Node]:
         node_ids = self._pos_node_map.get(pos, [])
 
-        return [self._nodes.get(node_id, None) for node_id in node_ids]
+        return [self._nodes_dict.get(node_id, None) for node_id in node_ids]
 
     def nodes_at(self, points: List[IVector]) -> Dict[IVector, List[Node]]:
         return {point: self.nodes_at_point(point) for point in points}
@@ -400,7 +457,7 @@ class Graph(object):
 
         edge_id = self._pos_edge_map.get(start, {}).get(end, None)
         if edge_id:
-            return self._edges.get(edge_id, None)
+            return self._edges_dict.get(edge_id, None)
 
         return None
 
@@ -411,10 +468,10 @@ class Graph(object):
 
     def __str__(self):
         res = "vertices: "
-        for id, node in self._nodes.items():
+        for id, node in self._nodes_dict.items():
             res += f"\n\t{str(node)}"
         res += "\nedges: "
-        for edge in self.edges():
+        for edge in self.edges:
             res += f"\n\t{str(edge)}"
         return res
 
@@ -431,7 +488,7 @@ class Graph(object):
             in graph """
         if path is None:
             path = []
-        nodes = self._nodes
+        nodes = self._nodes_dict
         path = path + [start_vertex]
         if start_vertex == end_vertex:
             return path
@@ -451,7 +508,7 @@ class Graph(object):
             end_vertex in graph """
         if path is None:
             path = []
-        nodes = self._nodes
+        nodes = self._nodes_dict
         path = path + [start_vertex]
         if start_vertex == end_vertex:
             return [path]
@@ -459,7 +516,7 @@ class Graph(object):
             return []
         paths = []
         for edge_id in self._node_edge_map[start_vertex]:
-            edge = self._edges[edge_id]
+            edge = self._edges_dict[edge_id]
             node = edge.end
             if node not in path:
                 extended_paths = self.find_all_paths(node,
@@ -476,7 +533,7 @@ class Graph(object):
             of adjacent vertices. """
 
         edges = self._node_edge_map.get(node, [])
-        edges = [self._edges[x] for x in edges]
+        edges = [self._edges_dict[x] for x in edges]
         degree = len(edges) + edges.count([x for x in edges if x.end == node])
         return degree
 
@@ -484,7 +541,7 @@ class Graph(object):
     def degree_sequence(self):
         """ calculates the degree sequence """
         seq = []
-        for id, node in self._nodes.items():
+        for id, node in self._nodes_dict.items():
             seq.append(self.vertex_degree(node))
         seq.sort(reverse=True)
         return tuple(seq)
@@ -503,7 +560,7 @@ class Graph(object):
     def delta(self):
         """ the minimum degree of the vertices """
         min = 100000000
-        for id, node in self._nodes.items():
+        for id, node in self._nodes_dict.items():
             vertex_degree = self.vertex_degree(node)
             if vertex_degree < min:
                 min = vertex_degree
@@ -513,7 +570,7 @@ class Graph(object):
     def Delta(self):
         """ the maximum degree of the vertices """
         max = 0
-        for id, node in self._nodes.items():
+        for id, node in self._nodes_dict.items():
             vertex_degree = self.vertex_degree(node)
             if vertex_degree > max:
                 max = vertex_degree
@@ -522,7 +579,7 @@ class Graph(object):
 
     def density(self):
         """ method to calculate the density of a graph """
-        g = self._nodes
+        g = self._nodes_dict
         V = len(g.keys())
         E = len(self.edges())
         return 2.0 * E / (V * (V - 1))
@@ -566,7 +623,12 @@ class Graph(object):
             return False
         return True
 
-    def astar(self, start: Node, end: Node, g_func: Callable[[Node, Node], float] = None, h_func: Callable[[Node, Node], float] = None, ignored_disablers:List[str]=None) -> AStarResults:
+    def astar(self,
+              start: Node,
+              end: Node,
+              g_func: Callable[[Node, Node], float] = None,
+              h_func: Callable[[Node, Node], float] = None,
+              ignored_disablers:List[str]=None) -> AStarResults:
 
         if not ignored_disablers:
             ignored_disablers = []
@@ -675,26 +737,25 @@ class Graph(object):
     def node_by_name(self, node_name: str):
         # nodes = self.nodes()
         # return next(node for node in nodes if node.name == node_name)
-        return self._nodes[self._node_by_name_map[node_name]]
+        return self._nodes_dict[self._node_by_name_map[node_name]]
 
 
 
     def verify_edge_configuration(self, edges_to_compare: List[Edge]):
         for edge in edges_to_compare:
-            if not self._edges[self._pos_edge_map[edge.start.pos][edge.end.pos]].config_match(edge):
+            if not self._edges_dict[self._pos_edge_map[edge.start.pos][edge.end.pos]].config_match(edge):
                 return False
 
         return True
 
-    '''A recursive function that find articulation points  
-        using DFS traversal 
-        u --> The vertex to be visited next 
-        visited[] --> keeps tract of visited vertices 
-        disc[] --> Stores discovery times of visited vertices 
-        parent[] --> Stores parent vertices in DFS tree 
-        ap[] --> Store articulation points'''
-
     def APUtil(self, u, visited, ap, parent, low, disc, iter):
+        '''A recursive function that find articulation points
+            using DFS traversal
+            u --> The vertex to be visited next
+            visited[] --> keeps tract of visited vertices
+            disc[] --> Stores discovery times of visited vertices
+            parent[] --> Stores parent vertices in DFS tree
+            ap[] --> Store articulation points'''
 
         # Count of children in current node
         children = 0
@@ -707,6 +768,9 @@ class Graph(object):
         low[u] = iter
         new_iter = iter + 1
 
+        if u.pos.as_tuple() == (960, 450):
+            deb = True
+
         # Recur for all the vertices adjacent to this vertex
         for v in self._graph_dict[u]:
             # If v is not visited yet, then make it a child of u
@@ -723,12 +787,12 @@ class Graph(object):
                 # u is an articulation point in following cases
                 # (1) u is root of DFS tree and has two or more chilren.
                 if parent[u] == -1 and children > 1:
-                    ap[u] = True
+                    ap[u] = children
 
                 # (2) If u is not root and low value of one of its child is more
                 # than discovery value of u.
                 if parent[u] != -1 and low[v] >= disc[u]:
-                    ap[u] = True
+                    ap[u] = children
 
                     # Update low value of u for parent function calls
             elif v != parent[u]:
@@ -736,22 +800,17 @@ class Graph(object):
 
                 # The function to do DFS traversal. It uses recursive APUtil()
 
-    def AP(self):
+    def AP(self) -> Dict[Node, int]:
+        # https://www.geeksforgeeks.org/articulation-points-or-cut-vertices-in-a-graph/
 
         # Mark all the vertices as not visited
         # and Initialize parent and visited,
         # and ap(articulation point) arrays
-        visited = {node: False for id, node in self._nodes.items()}
+        visited = {node: False for id, node in self._nodes_dict.items()}
         disc = {node: float("Inf") for node in self._graph_dict.keys()}
         low = {node: float("Inf") for node in self._graph_dict.keys()}
         parent = {node: -1 for node in self._graph_dict.keys()}
-        ap = {node: False for node in self._graph_dict.keys()}  # To store articulation points
-
-        print(visited)
-        print(disc)
-        print(low)
-        print(parent)
-        print(ap)
+        ap = {node: 0 for node in self._graph_dict.keys()}  # To store articulation points
 
         # Call the recursive helper function
         # to find articulation points
@@ -760,15 +819,23 @@ class Graph(object):
             if visited[i] == False:
                 self.APUtil(i, visited, ap, parent, low, disc, 0)
 
-        print(f"TESTTEST ap{ap}")
-        for index, value in enumerate(ap):
-            if value == True:
-                print(index)
+        ret = {k: v for k, v in ap.items() if v > 0}
+
+        logging.debug(f"Articulation points evaluation:"
+                      f"\n\tVisited: {visited}"
+                      f"\n\tdisc: {disc}"
+                      f"\n\tlow: {low}"
+                      f"\n\tparent: {parent}"
+                      f"\n\tap: {ap}"
+                      f"\nAPs: [{[k for k, v in ret.items()]}]")
+
+        return ret
+
 
     def closest_nodes(self, pos: IVector) -> List[Node]:
         closest_nodes = None
         closest_distance = None
-        for node in self.nodes():
+        for node in self.nodes:
             distance = node.pos.distance_from(pos)
             if closest_nodes is None or distance < closest_distance:
                 closest_nodes = [node]
@@ -780,9 +847,34 @@ class Graph(object):
 
         return closest_nodes
 
+
+    #TODO: enable a "walk" method. should accept a number of steps, an optional start point, and and optional end point
+
+    # def walk(self, name: str, points: [(int, int)]) -> ({str, Node}, {str, Edge}):
+    #     walk_nodes: {str, Node} = {}
+    #     walk_edges: {str, Edge} = {}
+    #
+    #     old_point = None
+    #     for point in points:
+    #         new_point = Node(self.nodeIndex, point[0], point[1])
+    #         walk_nodes[str(self.nodeIndex)] = new_point
+    #         self.nodeIndex += 1
+    #         if old_point is not None:
+    #             walk_edges[str(self.edgeIndex)] = Edge(self.edgeIndex, old_point, new_point, EdgeConnection.OneWay)
+    #             self.edgeIndex += 1
+    #
+    #         old_point = new_point
+    #
+    #     self.nodes.update(walk_nodes)
+    #     self.edges.update(walk_edges)
+    #     ret = (walk_nodes, walk_edges)
+    #     self.walks[name] = ret
+    #     return ret
+
+
     def copy(self):
         copy = Graph(graph_dict=self._graph_dict)
-        for edge in copy.edges():
+        for edge in copy.edges:
             og_edge = self.edge_between(edge.start, edge.end)
             for disabler in og_edge.disablers():
                 edge.add_disabler(disabler)
@@ -791,6 +883,10 @@ class Graph(object):
 
 if __name__ == "__main__":
     from coopstructs.vectors import Vector2
+    import logging
+
+    logging.basicConfig(level=logging.DEBUG)
+
     a = Node(name='A', pos=Vector2(0, 0))
     b = Node(name='B', pos=Vector2(3, 3))
     c = Node(name='C', pos=Vector2(2, 0))
@@ -807,4 +903,9 @@ if __name__ == "__main__":
          }
 
     graph = Graph(g)
-    print(graph.find_isolated_vertices())
\ No newline at end of file
+
+    #TEST 1
+    # print(graph.find_isolated_vertices())
+
+    #TEST 2
+    print(graph.AP())
\ No newline at end of file
